Imports System.Collections

Namespace NistRandomness
  Friend Class NistProgram
	Shared Sub Main(ByVal args() As String)
	  Try
		Console.WriteLine(vbCrLf & "Begin NIST tests of randomness using VB demo" & vbCrLf)

                'Dim ig As Double = GammaFunctions.GammaUpper(600, 600) ' see TS p. E-2. Maple = 0.4945710333, Matlab = 0.4945710331, TS = 0.4945710331
                'Console.WriteLine(ig) ' adjust desired accuracy, max-iterations -> VB = 0.4945710332

                'Dim bitString As String = "1100 1001 0000 1111 1101 1010 1010 0010 0010 0001 0110 1000 1100 0010 0011 0100 1100 0100 1100 0110 0110 0010 1000 1011 1000" ' page 2-5
		Dim bitString As String = "1100 0011 1010 1110 0000 1111 0000 1111 0000 1111 0000 1111 0000"

		Dim bitArray As BitArray = MakeBitArray(bitString)
		Console.WriteLine("Input sequence to test for randomness: " & vbCrLf)
		ShowBitArray(bitArray, 4, 52)

		Console.WriteLine(vbCrLf & vbCrLf & "1. Testing input frequencies")
		Dim pFreq As Double = FrequencyTest(bitArray)
		Console.WriteLine("pValue for Frequency test = " & pFreq.ToString("F4"))
		If pFreq < 0.01 Then
		  Console.WriteLine("There is evidence that sequence is NOT random")
		Else
		  Console.WriteLine("Sequence passes NIST frequency test for randomness")
		End If

		Dim blockLength As Integer = 8
		Console.WriteLine(vbCrLf & vbCrLf & "2. Testing input blocks (block length = " & blockLength & ")")
		Dim pBlock As Double = BlockTest(bitArray, blockLength)
		Console.WriteLine("pValue for Block test = " & pBlock.ToString("F4"))
		If pBlock < 0.01 Then
		  Console.WriteLine("There is evidence that sequence is NOT random")
		Else
		  Console.WriteLine("Sequence passes NIST block test for randomness")
		End If

		Console.WriteLine(vbCrLf & vbCrLf & "3. Testing input runs")
		Dim pRuns As Double = RunsTest(bitArray)
		Console.WriteLine("pValue for Runs test = " & pRuns.ToString("F4"))
		If pRuns < 0.01 Then
		  Console.WriteLine("There is evidence that sequence is NOT random")
		Else
		  Console.WriteLine("Sequence passes NIST runs test for randomness")
		End If

		Console.WriteLine(vbCrLf & vbCrLf & "End NIST randomness demo" & vbCrLf)
		Console.ReadLine()
	  Catch ex As Exception
		Console.WriteLine(ex.Message)
		Console.ReadLine()
	  End Try
	End Sub ' Main

	' ------------------------------------------------------------------

	Private Shared Function MakeBitArray(ByVal bitString As String) As BitArray
	  ' ex: string "010 101" -> a BitArray of false,true,false,true,false,true
	  Dim size As Integer = 0
	  For i  = 0 To bitString.Length - 1
		If bitString.Chars(i) <> " "c Then
			size += 1
		End If
	  Next i

	  Dim result As New BitArray(size) ' default is false
	  Dim k As Integer = 0 ' ptr into result
	  For i  = 0 To bitString.Length - 1
		If bitString.Chars(i) = " "c Then
			Continue For
		End If
		If bitString.Chars(i) = "1"c Then
		  result(k) = True
		Else
		  result(k) = False ' not necessary in VB
		End If
		k += 1
	  Next i
	  Return result
	End Function

	Private Shared Sub ShowBitArray(ByVal bitArray As BitArray, ByVal blockSize As Integer, ByVal lineSize As Integer)
	  For i  = 0 To bitArray.Length - 1
		If i > 0 AndAlso i Mod blockSize = 0 Then
		  Console.Write(" ")
		End If

		If i > 0 AndAlso i Mod lineSize = 0 Then
		  Console.WriteLine("")
		End If

		If bitArray(i) = False Then
			Console.Write("0")
		Else
			Console.Write("1")
		End If
	  Next i
	  Console.WriteLine("")
	End Sub

	' ------------------------------------------------------------------

	Private Shared Function FrequencyTest(ByVal bitArray As BitArray) As Double
	  ' perform a NIST frequency test on bitArray
	  Dim sum As Double = 0
	  For i  = 0 To bitArray.Length - 1
		If bitArray(i) = False Then
		  sum = sum - 1
		Else
		  sum = sum + 1
		End If
	  Next i
	  Dim testStat As Double = Math.Abs(sum) / Math.Sqrt(bitArray.Length)
	  Dim rootTwo As Double = 1.414213562373095
	  Dim pValue As Double = ErrorFunctionComplement(testStat / rootTwo)
	  Return pValue
	End Function

	Private Shared Function ErrorFunction(ByVal x As Double) As Double
	  ' assume x > 0.0
	  ' Abramowitz and Stegun eq. 7.1.26
	  Dim p As Double = 0.3275911
	  Dim a1 As Double = 0.254829592
	  Dim a2 As Double = -0.284496736
	  Dim a3 As Double = 1.421413741
	  Dim a4 As Double = -1.453152027
	  Dim a5 As Double = 1.061405429
	  Dim t As Double = 1.0 / (1.0 + p * x)
	  Dim err As Double = 1.0 - (((((a5 * t + a4) * t) + a3) * t + a2) * t + a1) * t * Math.Exp(-x * x)
	  Return err
	End Function

	Private Shared Function ErrorFunctionComplement(ByVal x As Double) As Double
	  Return 1 - ErrorFunction(x)
	End Function

	' ------------------------------------------------------------------

	Private Shared Function BlockTest(ByVal bitArray As BitArray, ByVal blockLength As Integer) As Double
	  ' NIST intra-block frequency test
	  Dim numBlocks As Integer = bitArray.Length \ blockLength ' 'N'

	  Dim proportions(numBlocks - 1) As Double
	  Dim k As Integer = 0 ' ptr into bitArray
	  For block  = 0 To numBlocks - 1
		Dim countOnes As Integer = 0
		For i  = 0 To blockLength - 1
		  Dim tempVar As boolean = bitArray(k) = True
		  k += 1
		  If tempVar Then
			countOnes += 1
		  End If
		Next i

		proportions(block) = (countOnes * 1.0) / blockLength
	  Next block

	  Dim summ As Double = 0.0
	  For block  = 0 To numBlocks - 1
		summ = summ + (proportions(block) - 0.5) * (proportions(block) - 0.5)
	  Next block
	  Dim chiSquared As Double = 4 * blockLength * summ

	  Dim a As Double = numBlocks / 2.0
	  Dim x As Double = chiSquared / 2.0
	  Dim pValue As Double = GammaFunctions.GammaUpper(a, x)
	  Return pValue
	End Function

	Private Shared Function RunsTest(ByVal bitArray As BitArray) As Double
	  ' NIST Runs test
	  Dim numOnes As Double = 0.0
	  For i  = 0 To bitArray.Length - 1
		If bitArray(i) = True Then
		  numOnes += 1
		End If
	  Next i

	  Dim prop As Double = (numOnes * 1.0) / bitArray.Length

            'Dim tau As Double = 2.0 / Math.Sqrt(bitArray.Length * 1.0)
            'If Math.Abs(prop - 0.5) >= tau Then
            '    Return 0.0 ' not-random short-circuit
            'End If

	  Dim runs As Integer = 1
	  For i  = 0 To bitArray.Length - 2
		If bitArray(i) <> bitArray(i + 1) Then
		  runs += 1
		End If
	  Next i

	  Dim num As Double = Math.Abs(runs - (2 * bitArray.Length * prop * (1 - prop)))
	  Dim denom As Double = 2 * Math.Sqrt(2.0 * bitArray.Length) * prop * (1 - prop)
	  Dim pValue As Double = ErrorFunctionComplement(num / denom)
	  Return pValue
	End Function

	' ------------------------------------------------------------------
  End Class ' Program

  Public Class GammaFunctions
	Public Shared Function GammaLower(ByVal a As Double, ByVal x As Double) As Double
	  ' incomplete Gamma 'P' (lower) aka 'igam'
	  If x < 0.0 OrElse a <= 0.0 Then
		Throw New Exception("Bad args in GammaLower")
	  End If
	  If x < a + 1 Then
		Return GammaLowerSer(a, x) ' no surprise
	  Else
		Return 1.0 - GammaUpperCont(a, x) ' indirectly is faster
	  End If
	End Function

	Public Shared Function GammaUpper(ByVal a As Double, ByVal x As Double) As Double
	  ' incomplete Gamma 'Q' (upper) == (1 - GammaP) but computed for efficiency
	  ' aka 'igamc' (incomplete gamma complement)
	  If x < 0.0 OrElse a <= 0.0 Then
		Throw New Exception("Bad args in GammaUpper")
	  End If
	  If x < a + 1 Then
		Return 1.0 - GammaLowerSer(a, x) ' indirect is faster
	  Else
		Return GammaUpperCont(a, x)
	  End If
	End Function

	' -------------------------------------------------------------------------------

	Private Shared Function LogGamma(ByVal x As Double) As Double
	  ' Log of Gamma from Lanczos with g=5, n=6/7
	  ' not in A & S 
	  Dim coef() As Double = { 76.18009172947146, -86.50532032941677, 24.01409824083091, -1.231739572450155, 0.1208650973866179E-2, -0.5395239384953E-5 }
	  Dim LogSqrtTwoPi As Double = 0.91893853320467274178
	  Dim denom As Double = x + 1
	  Dim y As Double = x + 5.5
	  Dim series As Double = 1.000000000190015
	  For i  = 0 To 5
		series += coef(i) / denom
		denom += 1.0
	  Next i
	  Return (LogSqrtTwoPi + (x + 0.5) * Math.Log(y) - y + Math.Log(series / x))
	End Function

	Private Shared Function GammaLowerSer(ByVal a As Double, ByVal x As Double) As Double
	  ' incomplete gamma lower (computed by series expansion)
	  If x < 0.0 Then
		Throw New Exception("x param less than 0.0 in GammaLowerSer")
	  End If

	  Dim gln As Double = LogGamma(a)
	  Dim ap As Double = a
	  Dim del As Double = 1.0 / a
	  Dim sum As Double = del
	  For n  = 1 To 100
		ap += 1
		del *= x / ap
		sum += del
		If Math.Abs(del) < Math.Abs(sum) * 3.0E-7 Then ' close enough?
		  Return sum * Math.Exp(-x + a * Math.Log(x) - gln)
		End If
	  Next n
	  Throw New Exception("Unable to compute GammaLowerSer to desired accuracy")
	End Function

	Private Shared Function GammaUpperCont(ByVal a As Double, ByVal x As Double) As Double
	  ' incomplete gamma upper computed by continuing fraction
	  If x < 0.0 Then
		Throw New Exception("x param less than 0.0 in GammaUpperCont")
	  End If
	  Dim gln As Double = LogGamma(a)
	  Dim b As Double = x + 1.0 - a
	  Dim c As Double = 1.0 / 1.0E-30 ' div by close to double.MinValue
	  Dim d As Double = 1.0 / b
	  Dim h As Double = d
	  For i  = 1 To 100
		Dim an As Double = -i * (i - a)
		b += 2.0
		d = an * d + b
		If Math.Abs(d) < 1.0E-30 Then ' got too small?
			d = 1.0E-30
		End If
		c = b + an / c
		If Math.Abs(c) < 1.0E-30 Then
			c = 1.0E-30
		End If
		d = 1.0 / d
		Dim del As Double = d * c
		h *= del
		If Math.Abs(del - 1.0) < 3.0E-7 Then
		  Return Math.Exp(-x + a * Math.Log(x) - gln) * h ' close enough?
		End If
	  Next i
	  Throw New Exception("Unable to compute GammaUpperCont to desired accuracy")
	End Function

	' -------------------------------------------------------------------------------

        'Public Shared Function LogGamma(ByVal z As Double) As Double
        '    '  alternate, not quite as accurate, LogGamma()
        '    '   A & S 6.1.41 (Stirling's approximation)
        '    Dim x1 As Double = (z - 0.5) * Math.Log(z)
        '    Dim x3 As Double = 0.5 * Math.Log(2 * Math.PI)
        '    Dim x4 As Double = 1 / (12 * z)
        '    Dim x5 As Double = 1 / (360 * z * z * z)
        '    Dim x6 As Double = 1 / (1260 * z * z * z * z * z)
        '    Dim x7 As Double = 1 / (1680 * z * z * z * z * z * z * z)
        '    Return x1 - z + x3 + x4 - x5 + x6 - x7
        'End Function
  End Class ' GammaFunctions
End Namespace ' ns
